# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"detuplify_pg": "01_optimizer.ipynb",
         "set_item_pg": "01_optimizer.ipynb",
         "pytorch_hp_map": "01_optimizer.ipynb",
         "OptimWrapper": "01_optimizer.ipynb",
         "Adam": "01_optimizer.ipynb",
         "SGD": "01_optimizer.ipynb",
         "params": "01_optimizer.ipynb",
         "convert_params": "01_optimizer.ipynb",
         "event.__module__": "02_callback.core.ipynb",
         "Callback": "02_callback.core.ipynb",
         "TrainEvalCallback": "02_callback.core.ipynb",
         "GatherPredsCallback": "02_callback.core.ipynb",
         "FetchPredsCallback": "02_callback.core.ipynb",
         "DataLoaders": "03_learner.ipynb",
         "defaults.lr": "03_learner.ipynb",
         "replacing_yield": "03_learner.ipynb",
         "mk_metric": "03_learner.ipynb",
         "save_model": "03_learner.ipynb",
         "load_model": "03_learner.ipynb",
         "Learner": "03_learner.ipynb",
         "before_batch_cb": "03_learner.ipynb",
         "Learner.save": "03_learner.ipynb",
         "Learner.load": "03_learner.ipynb",
         "Learner.export": "03_learner.ipynb",
         "load_learner": "03_learner.ipynb",
         "to_detach_from_dl": "03_learner.ipynb",
         "Metric": "03_learner.ipynb",
         "AvgMetric": "03_learner.ipynb",
         "AvgLoss": "03_learner.ipynb",
         "AvgSmoothLoss": "03_learner.ipynb",
         "ValueMetric": "03_learner.ipynb",
         "Recorder": "03_learner.ipynb",
         "Learner.freeze_to": "03_learner.ipynb",
         "Learner.freeze": "03_learner.ipynb",
         "Learner.unfreeze": "03_learner.ipynb",
         "Learner.tta": "03_learner.ipynb",
         "flatten_check": "04_metrics.ipynb",
         "AccumMetric": "04_metrics.ipynb",
         "skm_to_fastai": "04_metrics.ipynb",
         "optim_metric": "04_metrics.ipynb",
         "accuracy": "04_metrics.ipynb",
         "error_rate": "04_metrics.ipynb",
         "top_k_accuracy": "04_metrics.ipynb",
         "APScoreBinary": "04_metrics.ipynb",
         "BalancedAccuracy": "04_metrics.ipynb",
         "BrierScore": "04_metrics.ipynb",
         "CohenKappa": "04_metrics.ipynb",
         "F1Score": "04_metrics.ipynb",
         "FBeta": "04_metrics.ipynb",
         "HammingLoss": "04_metrics.ipynb",
         "Jaccard": "04_metrics.ipynb",
         "Precision": "04_metrics.ipynb",
         "Recall": "04_metrics.ipynb",
         "RocAuc": "04_metrics.ipynb",
         "RocAucBinary": "04_metrics.ipynb",
         "MatthewsCorrCoef": "04_metrics.ipynb",
         "Perplexity": "04_metrics.ipynb",
         "perplexity": "04_metrics.ipynb",
         "accuracy_multi": "04_metrics.ipynb",
         "APScoreMulti": "04_metrics.ipynb",
         "BrierScoreMulti": "04_metrics.ipynb",
         "F1ScoreMulti": "04_metrics.ipynb",
         "FBetaMulti": "04_metrics.ipynb",
         "HammingLossMulti": "04_metrics.ipynb",
         "JaccardMulti": "04_metrics.ipynb",
         "MatthewsCorrCoefMulti": "04_metrics.ipynb",
         "PrecisionMulti": "04_metrics.ipynb",
         "RecallMulti": "04_metrics.ipynb",
         "RocAucMulti": "04_metrics.ipynb",
         "mse": "04_metrics.ipynb",
         "rmse": "04_metrics.ipynb",
         "rmse.__doc__": "04_metrics.ipynb",
         "mae": "04_metrics.ipynb",
         "msle": "04_metrics.ipynb",
         "exp_rmspe": "04_metrics.ipynb",
         "exp_rmspe.__doc__": "04_metrics.ipynb",
         "ExplainedVariance": "04_metrics.ipynb",
         "R2Score": "04_metrics.ipynb",
         "PearsonCorrCoef": "04_metrics.ipynb",
         "SpearmanCorrCoef": "04_metrics.ipynb",
         "foreground_acc": "04_metrics.ipynb",
         "Dice": "04_metrics.ipynb",
         "DiceMulti": "04_metrics.ipynb",
         "JaccardCoeff": "04_metrics.ipynb",
         "CorpusBLEUMetric": "04_metrics.ipynb",
         "LossMetric": "04_metrics.ipynb",
         "LossMetrics": "04_metrics.ipynb",
         "ProgressCallback": "05_callback.training_utils.ipynb",
         "Learner.no_bar": "05_callback.training_utils.ipynb",
         "CollectDataCallback": "05_callback.training_utils.ipynb",
         "CudaCallback": "05_callback.training_utils.ipynb",
         "annealer": "06_callback.training.ipynb",
         "sched_lin": "06_callback.training.ipynb",
         "sched_cos": "06_callback.training.ipynb",
         "sched_no": "06_callback.training.ipynb",
         "sched_exp": "06_callback.training.ipynb",
         "SchedLin": "06_callback.training.ipynb",
         "SchedCos": "06_callback.training.ipynb",
         "SchedNo": "06_callback.training.ipynb",
         "SchedExp": "06_callback.training.ipynb",
         "SchedPoly": "06_callback.training.ipynb",
         "combine_scheds": "06_callback.training.ipynb",
         "combined_cos": "06_callback.training.ipynb",
         "ParamScheduler": "06_callback.training.ipynb",
         "Learner.fit_one_cycle": "06_callback.training.ipynb",
         "Recorder.plot_sched": "06_callback.training.ipynb",
         "Learner.fit_flat_cos": "06_callback.training.ipynb",
         "Learner.fit_sgdr": "06_callback.training.ipynb",
         "Learner.fine_tune": "06_callback.training.ipynb",
         "LRFinder": "06_callback.training.ipynb",
         "Recorder.plot_lr_find": "06_callback.training.ipynb",
         "SuggestedLRs": "06_callback.training.ipynb",
         "Learner.lr_find": "06_callback.training.ipynb"}

modules = ["optimizer.py",
           "callback/core.py",
           "learner.py",
           "metrics.py",
           "callback/training_utils.py",
           "callback/training.py"]

doc_url = "https://muellerzr.github.io/fastai_minima/"

git_url = "https://github.com/muellerzr/fastai_minima/tree/master/"

def custom_doc_links(name): return None
